CHARDB_UPDATE_INTERVAL = .1; -- Tenth of a second
CHARDB_UNITCACHE_CHECK_INTERVAL = 60*60; -- Ten minutes
CHARDB_INSPECT_TIMEOUT = 4; -- Four seconds

CHARDB_DEBUG = false;

CHARDB_TALENT_TAB_NAMES = {};

CharDB = {};
CharDB.EventHandlers = {};
CharDB.timeSinceLastUpdate = 0;
CharDB.runningUpdate = false;
CharDB.output = nil;
CharDB.button = nil;
CharDB.inspectee = nil;
CharDB.inspecteeGUID = nil;
CharDB.inspectTimeout = 0;
CharDB.inspectQueue = {};
CharDB.printQueue = {};
CharDB.inspecting = false;
CharDB.timeSinceLastPlayerRefresh = CHARDB_UNITCACHE_CHECK_INTERVAL+1; -- Lets force a refresh
CharDBData = {}

-- Called at file end
function CharDB.Init()
	--Output
	CharDB.output = CreateFrame("EditBox", "CharDbOutput", UIParent);
	CharDB.output:SetAutoFocus(false);
 	CharDB.output:Hide();

	CharDB.button = CreateFrame("Button", "CharDbButton", Minimap);

 	-- Setup slash
 	SlashCmdList["CharDB"] = CharDB.SlashHandler;
 	SLASH_CharDB1 = "/cdb";
 	SLASH_CharDB2 = "/CharDB";

	-- Inspect check
	hooksecurefunc("NotifyInspect", function(target)
		CharDB.inspectee = target; 
		CharDB.inspecteeGUID = UnitGUID(target);
		CharDB.inspectTimeout = GetTime() + CHARDB_INSPECT_TIMEOUT;
		CharDB.inspecting = true;
	end)

	-- Event handling
	CharDB.output:RegisterEvent("ADDON_LOADED");
	CharDB.output:RegisterEvent("INSPECT_TALENT_READY");
	CharDB.output:RegisterEvent("PARTY_MEMBERS_CHANGED");
	CharDB.output:RegisterEvent("RAID_ROSTER_UPDATE");

	CharDB.output:SetScript("OnEvent", function(self, event, ...)
	  CharDB.EventHandlers[event](...);
	end)
	
	CharDB.button:SetScript("OnUpdate", function(self, elapsed)
	  CharDB.EventHandlers.OnUpdate(elapsed);
	end)
	
end

-- Call after all saved variables are loaded
function CharDB.EventHandlers.ADDON_LOADED(addon) 
	if (addon ~= "CharDB") then return; end
	
	if (CharDBData.minimapAngle == nil) then
		CharDBData.minimapAngle = 100
	end
	if (CharDBData.players == nil) then
		CharDBData.players = {}
	end
	if (CharDBData.gemType == nil) then
		CharDBData.gemType = 'epic'
	end
	
 	LibItemDB.Init();
 	CharDB.output:SetFrameStrata("BACKGROUND");
	CharDB.output:SetFont("Fonts\\FRIZQT__.TTF", 13, "");
	CharDB.output:SetTextColor(0, 0, 0, 1);
	CharDB.output:SetWidth(800);
	CharDB.output:SetHeight(50);
	CharDB.output:SetBackdropColor(1, 1, 1, 1);
	CharDB.output:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 32,
  	})
	CharDB.output:EnableMouse(true);
	CharDB.output:SetMovable(true);
	
	CharDB.output:SetClampedToScreen(true);

	CharDB.output:SetPoint("CENTER", 0, 100);
	CharDB.output:SetScript("OnEscapePressed", function() CharDB.output:ClearFocus(); end);
	CharDB.output:SetScript("OnMouseUp", function(self, button)
		if (button == 'RightButton') then CharDB.output:Hide();end
	end);
	
	local time = GetTime();
	
	--local item = nil
	
	GameTooltip:HookScript("OnTooltipSetUnit", function(tooltip)
		CharDB.compareFrame:Hide();
	end)
	GameTooltip:HookScript("OnTooltipSetSpell", function(tooltip)
		CharDB.compareFrame:Hide();
	end)
	
	GameTooltip:HookScript("OnTooltipSetItem", function(tooltip) 
		--local nowTime = GetTime();
		--if (time > nowTime) then return end;
		--time = nowTime + 0.5;
		local itemName, link = tooltip:GetItem();
		if (link == item) then return end
		item = link
		--Spew('item', item)
		local count = GetNumPartyMembers();
		local players = {CharDB.GetPlayerByUnitId("player")}
		for i=1,count do
			tinsert(players, CharDB.GetPlayerByUnitId("party"..i));
		end
	
		count = GetNumRaidMembers();
		for i=1,count do
			tinsert(players, CharDB.GetPlayerByUnitId("raid"..i))
		end
		
		local result = format('Item: %s\n', itemName);
		
		for _, player in ipairs(players) do
			local compare = player:CompareItem(link)
			local padding = strrep('  ', strlen(player.name))
--			Spew('padding', '['..padding..']')
--			Spew('name  2', '['..player.name..']')
			local second = false;
			for item, diff in pairs(compare) do
				local itemName = GetItemInfo(item)
				local formatString = '%s: %.1f (%s)\n';
				if(second) then
					formatString = '%4$s  %2$.1f (%3$s)\n'
				end
				result = result..format(formatString, player.name, diff, itemName, padding)
				second = true
			end
		end
		CharDB.compareFrameText:SetText(result)
		
		if (CharDB.compareFrame:GetRight() and CharDB.compareFrame:GetRight() and CharDB.compareFrameText:GetHeight()) then 
			CharDB.compareFrameText:SetPoint("TOPLEFT",CharDB.compareFrame,"TOPLEFT",10,-5)
			CharDB.compareFrameText:SetWidth(CharDB.compareFrameText:GetStringWidth() + 10)
			CharDB.compareFrame:SetHeight(CharDB.compareFrameText:GetHeight() + 10)
			CharDB.compareFrame:SetWidth(CharDB.compareFrameText:GetWidth() + 5)
		end
		
		CharDB.compareFrame:SetPoint('TOPLEFT', GameTooltip, 'TOPRIGHT', 10, 0)
		CharDB.compareFrame:Show();		
	end)
	
	GameTooltip:HookScript("OnShow", function(tooltip) 
		if (tooltip:GetItem() ~= nil) then
			CharDB.compareFrame:Show();
		end
	end)
	
	GameTooltip:HookScript("OnHide", function(tooltip) 
		CharDB.compareFrame:Hide();
	end)
	
	CharDB.compareFrame = CreateFrame('Frame', 'CharDB_CompareFrame', nil, 'GameTooltipTemplate');
	local compare = CharDB.compareFrame;
	compare:SetPoint('TOPLEFT', GameTooltip, 'TOPRIGHT', 10, 0)
	compare:SetWidth(300)
	compare:SetHeight(60)
	CharDB.compareFrameText = compare:CreateFontString('CharDB_CompareFrameText');
	local compareText = CharDB.compareFrameText;
	compareText:SetPoint("TOPLEFT",compare,"TOPLEFT",5,-5)
	compareText:SetPoint("TOPRIGHT",compare,"TOPRIGHT",-5,-5)
	compareText:SetFont("Fonts\\FRIZQT__.TTF", 10);
	compareText:SetJustifyH('LEFT')
	compare:Hide();
--	CharDB.button:SetMovable(true)
--	CharDB.button:EnableMouse(true)
--	CharDB.button:SetScript("OnMouseDown",function(self, button)
--		CharDB.button:StartMoving()
--	end)
--	CharDB.button:SetScript("OnMouseUp",function(self, button)
--		CharDB.button:StopMovingOrSizing()
--	end)
	
	CharDB.button:SetWidth(33);
	CharDB.button:SetHeight(33);
	CharDB.button:SetPoint("TOPLEFT");
	local button = CharDB.button;
	local background = button:CreateTexture('CharDBArtwork', "ARTWORK");
	background:SetTexture("Interface\\Icons\\Ability_Warrior_StalwartProtector");
	background:SetWidth(30)
	background:SetHeight(30)
	background:SetTexCoordModifiesRect(true)
	background:SetTexCoord(0, .6, 0, .6)
	background:SetPoint("TOPLEFT", button, 7, 6)
	local overlay = button:CreateTexture('CharDBBorder', "OVERLAY");
	overlay:SetTexture("Interface\\Minimap\\MiniMap-TrackingBorder");
	overlay:SetWidth(56)
	overlay:SetHeight(56)
	overlay:SetPoint("TOPLEFT", button)
	button:RegisterForClicks("LeftButtonUp","RightButtonUp")
	button:RegisterForDrag("RightButton")
	button:SetScript("OnEnter", function(self)
		ToggleDropDownMenu(1, nil, CharDB.dropdownMenu, self, -110, 10)
	end)
	button:SetScript("OnDragStart", function(self)
		if (IsShiftKeyDown()) then
			this:LockHighlight()
			CharDbButton_DraggingFrame:Show()
		end
	end)
	button:SetScript("OnDragStop", function(self)
		this:UnlockHighlight()
		CharDbButton_DraggingFrame:Hide()
	end)
	local draggingFrame = CreateFrame("Frame", "CharDbButton_DraggingFrame", button)
	draggingFrame:Hide();
	draggingFrame:SetScript("OnUpdate", function(self, dur) 
		local xpos,ypos = GetCursorPosition()
		local xmin,ymin = Minimap:GetLeft(), Minimap:GetBottom()

		xpos = xmin-xpos/UIParent:GetScale()+70 -- get coordinates as differences from the center of the minimap
		ypos = ypos/UIParent:GetScale()-ymin-70

		CharDBData.minimapAngle = math.deg(math.atan2(ypos,xpos)) -- save the degrees we are relative to the minimap center
		CharDB.button:SetPoint("TOPLEFT","Minimap","TOPLEFT",52-(80*cos(CharDBData.minimapAngle)),(80*sin(CharDBData.minimapAngle))-52)
	end)
--	CharDB.button:SetBackdropColor(1, 1, 1, 1);
--	CharDB.button:SetBackdrop({
--		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 32,
-- 	})
  	
  	CharDB.button:SetPoint("TOPLEFT","Minimap","TOPLEFT",52-(80*cos(CharDBData.minimapAngle)),(80*sin(CharDBData.minimapAngle))-52)
	CharDB.button:SetClampedToScreen(true);
	CharDB.button:Raise()
	CharDB.button:Show();
	
	local dropdownMenu = CreateFrame("Frame", "CharDB_MinimapButton_DropDownMenu")
	CharDB.dropdownMenu = dropdownMenu;
	dropdownMenu.displayMode = "MENU"
	
	local info = {}
	dropdownMenu.initialize = function(self, level)
		if not level then return end
		wipe(info)
		if level == 1 then
--	 		info.keepShownOnClick = 1
			info.disabled	 = nil
			info.isTitle	  = nil
			info.notCheckable = nil
			info.text = "Inspect";
			info.hasArrow = true;
			info.value = "inspect"
			UIDropDownMenu_AddButton(info, level)
			wipe(info)
	
--	 		info.keepShownOnClick = 1
			info.disabled = nil
			info.isTitle = nil
			info.notCheckable = nil
			info.text = "Simcraft";
			info.hasArrow = true;
			info.value = "simcraft"
			UIDropDownMenu_AddButton(info, level)
			wipe(info)
	
--	 		info.keepShownOnClick = 1
			info.disabled = nil
			info.isTitle = nil
			info.notCheckable = nil
			info.text = "Print";
			info.hasArrow = true;
			info.value = "print"
			UIDropDownMenu_AddButton(info, level)
			wipe(info)
	
			info.disabled = nil
			info.isTitle = nil
			info.notCheckable = true
			info.text = 'Show Output';
			info.hasArrow = false;
			info.func = function(self)
				CharDB.DoAction({['action']='show', ['target']=unitId, ['output']='self'});
			end
			UIDropDownMenu_AddButton(info, level)
			wipe(info)
	
			info.disabled = nil
			info.isTitle = nil
			info.notCheckable = true
			info.text = 'Hide Output';
			info.hasArrow = false;
			info.func = function(self)
				CharDB.DoAction({['action']='hide', ['target']=unitId, ['output']='self'});
			end
			UIDropDownMenu_AddButton(info, level)
			wipe(info)
			
			info.disabled = nil
			info.isTitle = nil
			info.notCheckable = true
			info.text = 'Parse Simcraft';
			info.hasArrow = false;
			info.func = function(self)
				CharDB.DoAction({['action']='parse_simcraft', ['input']='output'});
			end
			UIDropDownMenu_AddButton(info, level)
			wipe(info)
	
			-- Close menu item
--			info.text = CLOSE
--			info.func = function() CloseDropDownMenus() end
--			info.checked = nil
--			info.notCheckable = 1
--			UIDropDownMenu_AddButton(info, level)
		elseif level == 2 then
			for name, unitId in pairs({Player="player", Target="target", Focus="focus"}) do
				info.disabled	 = nil
				info.isTitle	  = nil
				info.notCheckable = true
				info.text = name;
				info.hasArrow = false;
				info.func = function(self, action, unitId)
					CharDB.DoAction({['action']=action, ['target']=unitId, ['output']='self'});
					CloseDropDownMenus()
				end
				info.arg1 = UIDROPDOWNMENU_MENU_VALUE
				info.arg2 = unitId
				UIDropDownMenu_AddButton(info, level)
				wipe(info)
			end
			
			info.disabled	 = nil
			info.isTitle	  = nil
			info.notCheckable = true
			info.text = 'Party';
			info.hasArrow = false;
			info.func = function(self, action)
				local t = {}
				for i=1,GetNumPartyMembers() do
					tinsert(t, "party"..i)
				end
				CharDB.DoAction({['action']=action, ['target']=t, ['output']='self'});
				CloseDropDownMenus()
			end
			info.arg1 = UIDROPDOWNMENU_MENU_VALUE
			UIDropDownMenu_AddButton(info, level)
			wipe(info)
			
			info.disabled	 = nil
			info.isTitle	  = nil
			info.notCheckable = true
			info.text = 'Raid';
			info.hasArrow = false;
			info.func = function(self, action)
				local t = {}
				for i=1,GetNumRaidMembers() do
					tinsert(t, "raid"..i)
				end
				CharDB.DoAction({['action']=action, ['target']=t, ['output']='self'});
				CloseDropDownMenus()
			end
			info.arg1 = UIDROPDOWNMENU_MENU_VALUE
			UIDropDownMenu_AddButton(info, level)
			wipe(info)
			
		end
	end
	
	for i, player in pairs(CharDBData.players) do
		setmetatable(player, CharDBUnit)
		player:ValidateUnitId();
	end
end

function CharDB.SlashHandler(msg)
	local rawArgs = {strsplit(" ", msg)};
	local options = {action="inspect", target="target", output="self"}
	for _, arg in pairs(rawArgs) do
		--DEFAULT_CHAT_FRAME:AddMessage(msg);
		--CharDB.Test = "Test: "..msg;
		if (arg == "show") then 
			options.action = "show";
		elseif (arg == "hide") then
			options.action = "hide";
		elseif (arg == "inspect") then
			options.action = "inspect";
		elseif (arg == "print") then
			options.action = "print";
		elseif (arg == "sim" or arg == "simcraft") then
			options.action = "simcraft";
		elseif (arg == "player") then
			options.target = "player";
		elseif (arg == "target") then
			options.target = "target";
		elseif (arg == "group") then
			local t = {}
			for i=1,GetNumPartyMembers() do
				tinsert(t, "party"..i)
			end
			options.target = t;
		elseif (arg == "raid") then
			local t = {}
			for i=1,GetNumRaidMembers() do
				tinsert(t, "raid"..i)
			end
			options.target = t;
		else
			CharDB.Print("Unknown argument: "..arg);
		end
	end
	CharDB.DoAction(options)
end

function CharDB.DoAction(actionTable)
	if (actionTable.action == "inspect") then
		CharDB.AddToInspectQueue(actionTable.target);
		CharDB.AddToPrintQueue(actionTable.target, actionTable.output, "normal");
	elseif (actionTable.action == "print") then
		CharDB.AddToPrintQueue(actionTable.target, actionTable.output, "normal")
	elseif (actionTable.action == "simcraft") then
		CharDB.AddToPrintQueue(actionTable.target, "frame", "simcraft")	
	elseif (actionTable.action == "show") then
		CharDB.output:Show()
	elseif (actionTable.action == "hide") then
		CharDB.output:Hide()
	elseif (actionTable.action == "parse_simcraft") then
		CharDB.ParseSimcraftOutput(CharDB.output:GetText())
	end
end

function CharDB.AddToInspectQueue(unitIds)
	if (type(unitIds) == "table") then
	elseif (type(unitIds) == "string") then
		unitIds = {unitIds}
	else
		CharDB.Debug(type(unitIds));
		assert(false);
	end
	for _, unitId in pairs(unitIds) do
		if (UnitIsPlayer(unitId)) then 
			local GUID = UnitGUID(unitId)
			if (CharDB.inspectQueue[GUID] ~= nil) then 
				CharDB.inspectQueue[GUID].unitId = unitId;
			else
				local player = CharDBData.players[GUID];
				if (player == nil) then
					player = CharDBUnit:New(unitId) 
					CharDBData.players[GUID] = player
				else 
					player.unitId = unitId;
					player:ResetGear();
					player:ResetTalents();
				end
				CharDB.Debug(UnitName(unitId).." added to inspect queue~") 
				CharDB.inspectQueue[GUID] = player
			end
		else
			CharDB.Debug(unitId.." is not valid unitid~")
		end
	end
end

function CharDB.AddToPrintQueue(unitIds, output, form)
	if (type(unitIds) == "table") then
	elseif (type(unitIds) == "string") then
		unitIds = {unitIds}
	else
		CharDB.Debug(type(unitIds));
		assert(false);
	end

	--	if (not UnitIsPlayer(unitId)) then return end
	local players = {}
	for _, unitId in pairs(unitIds) do	
		if (UnitIsPlayer(unitId)) then 
			local guid = UnitGUID(unitId)
			tinsert(players, guid)
			CharDB.Debug(UnitName(unitId).." added to print queue~") 
			local player = CharDBData.players[guid];
			if (player == nil or player:Outdated(time()-CHARDB_UNITCACHE_CHECK_INTERVAL)) then
				CharDB.AddToInspectQueue(unitId)
			end
		else
			CharDB.Debug(unitId.." is not valid unitid~")
		end
	end
	players.output = output;
	players.form = form;
	tinsert(CharDB.printQueue, players)
end

--[[
function CharDB.PrintData(unitIds, output, form) 
	if (type(unitIds) == "table") then
	elseif (type(unitIds) == "string") then
		unitIds = {unitIds}
	else
		assert(false);
	end
	
	local text = ""
	for _, unitId in pairs(unitIds) do
		if (CharDBData.players[UnitGUID(unitIdOrGuid)]) then
			text = text..CharDB.GetPlayerText(CharDBData.players[guid], form)..'\n';
		end
	end
	if (text ~= "") then
		CharDB.Print(text, playersGUID.output);
	end
end
]]

function CharDB.Debug(msg) 
	if (CHARDB_DEBUG) then CharDB.Print('CharDBDebug: '..msg, nil) end
end

function CharDB.GetPlayerText(playerObject, form)
	local text = nil;
	if (not playerObject:Complete()) then
		text = format("%s's data is still loading", playerObject.name)
	else
		if (form == "normal" or form == nil) then
			local spec = playerObject:GetSpec() 
			text = format("%s (%d/%d/%d) %s", playerObject.name, spec[1], spec[2], spec[3], playerObject.classLocal).."\n"
			for _, item in pairs(playerObject.gear) do
				text = text..item.."\n"
			end
		elseif (form == "simcraft" or form == "sim") then
			text = playerObject:GetSimcraftParams();
		else 
			CharDB.Debug("["..form.."] is a unknown form type!");
			return;
		end
	end
	assert(text ~= nil)
	return text;
end
	
function CharDB.PrintPlayerData(playerObject, output, form)
	local text = CharDB.GetPlayerText(playerObject, form);
	CharDB.Print(text, output)
end

function CharDB.Print(msg, output)
	if (output == nil or output == "self") then
		if (strfind(msg, '\n') ~= nil) then
			for _, line in ipairs({strsplit('\n', msg)}) do
				DEFAULT_CHAT_FRAME:AddMessage(line);
			end
		else
			DEFAULT_CHAT_FRAME:AddMessage(msg);
		end
	elseif (output == "frame") then
		CharDB.output:SetText(msg);
	 	CharDB.output:Show();
	else 
		CharDB.Debug("["..output.."] is a unknown output type!");
	end
end

function Print(msg) 
	DEFAULT_CHAT_FRAME:AddMessage(msg);
end

function CharDB.RunPeriodicTasks(elapsed) 
	CharDB.RunPlayerRefreshCheck(elapsed)
	CharDB.RunInspectQueue()
	CharDB.RunPrintQueue();
end

function CharDB.RunPlayerRefreshCheck(elapsed)
	CharDB.timeSinceLastPlayerRefresh = CharDB.timeSinceLastPlayerRefresh + elapsed;
	if (CharDB.timeSinceLastPlayerRefresh > CHARDB_UNITCACHE_CHECK_INTERVAL) then
		CharDB.Debug(format("Running player refresh, elapsed: %.3f~", CharDB.timeSinceLastPlayerRefresh));
		CharDB.timeSinceLastPlayerRefresh = 0;
		CharDB.AddToInspectQueue("player")
		local count = GetNumPartyMembers();
		for i=1,count do
			CharDB.AddToInspectQueue("party"..i)
		end
		count = GetNumRaidMembers();
		for i=1,count do
			CharDB.AddToInspectQueue("raid"..i)
		end
	end
end

function CharDB.RunInspectQueue()
	local time = GetTime()
	local player = nil;
	if (CharDB.inspectee ~= nil and CanInspect(CharDB.inspectee) and not UnitIsUnit(CharDB.inspectee, 'player')) then 		
		CharDB.InspectGear(CharDB.inspectee, CharDB.inspectTimeout > time);
	end

	for guid, player in pairs(CharDB.inspectQueue) do
		if (not player:ValidateUnitId()) then
			CharDB.Debug(player.name.." has an invalid unitid~")
			CharDB.inspectQueue[guid] = nil
		elseif (player:Complete()) then
			CharDB.Debug(player.name.." is complete~")
			CharDB.inspectQueue[guid] = nil
		elseif ((CharDB.inspectTimeout < time or CharDB.inspecting == false)
			and CanInspect(player.unitId)) then
			CharDB.Debug(player.name.." is being inspected~") 
			CharDB.Debug(format("Timeout:%.3f Time:%.3f Inspecting:%s~", CharDB.inspectTimeout, time, tostring(CharDB.inspecting)));
			NotifyInspect(player.unitId); -- inspectTimeout = time + INSPECT_TIMEOUT, inspecting = true
			return;
		end
	end
end

function CharDB.RunPrintQueue()
	local time = GetTime();
	local players, ready, player, text;
	for index, playersGUID in pairs(CharDB.printQueue) do
		players = {};
		ready = true;
		for i=1,#playersGUID do
			player = CharDBData.players[playersGUID[i]];
			if (player:Complete()) then
				tinsert(players, player);
			else
				ready = false;
				break;
			end
		end
		if (ready) then
			text = "";
			for _, player in pairs(players) do
				CharDB.Debug(player.name.." is being printed~") 
				text = text..CharDB.GetPlayerText(player, playersGUID.form)..'\n';
			end
			CharDB.Print(text, playersGUID.output);
			CharDB.printQueue[index] = nil;
		end
	end
end

function CharDB.EventHandlers.INSPECT_TALENT_READY()
	if CharDB.inspecteeGUID == UnitGUID(CharDB.inspectee) then -- Make sure inspectee hasn't changed
		local guid = CharDB.inspecteeGUID;
		if (CharDBData.players[guid] ~= nil) then
			local talentGroups = {};
			local player = CharDBData.players[guid];
		--	CharDB.Debug(player.name.." has inspect talents~") 
			local inspecting = not UnitIsUnit(CharDB.inspectee, "player");
			local tabCount = GetNumTalentTabs(inspecting);
			if(CHARDB_TALENT_TAB_NAMES[player.class] == nil and player.class ~= nil) then
				CHARDB_TALENT_TAB_NAMES[player.class] = {};
			end
			local talentGroupCount = GetNumTalentGroups(inspecting);
			for group = 1,talentGroupCount do
				talentGroups[group] = {};
				for tab = 1, tabCount do
					if(	player.class ~= nil
					   and CHARDB_TALENT_TAB_NAMES[player.class] ~= nil
					   and CHARDB_TALENT_TAB_NAMES[player.class][tab] == nil) then
						CHARDB_TALENT_TAB_NAMES[player.class][tab] = GetTalentTabInfo(tab, inspecting, nil, group);
					end
					talentGroups[group][tab] = {};
					local talentCount = GetNumTalents(tab, inspecting, nil, group);
					for talent = 1, talentCount do
						talentGroups[group][tab][talent] = select(5, GetTalentInfo(tab, talent, inspecting, nil, group));
					end
				end
			end
			
			local activeTalentGroup = GetActiveTalentGroup(inspecting);

			if CharDB.inspecteeGUID == UnitGUID(CharDB.inspectee) then -- Make sure inspectee hasn't changed
				player:SetTalents(talentGroups, activeTalentGroup)
			end
			
			CharDB.InspectGear(CharDB.inspectee, false)
		end
	end
	CharDB.inspecting = false;
end

function CharDB.InspectGear(unitId, forceCompletion)
	local guid = UnitGUID(unitId)
	local player = CharDBData.players[guid];
	if (player ~= nil) then
	--	CharDB.Debug(player.name.." has inspect gear~") 
		local complete = true;
		local gear = {};
		local gearPartial = false;
		local itemLink = nil;
		local itemData = nil
		for i = 1, 19 do
			itemLink = GetInventoryItemLink(unitId, i);
			if (itemLink) then
				itemData = {LibItemDB.SplitLink(itemLink)};
				--if  (itemData[3] == 0 and itemData[4] == 0
				-- and itemData[5] == 0 and itemData[6] == 0 and itemData[7] == 0  
				-- and (itemData[9] == 0 or (itemData[8] < 0 and bit.mod(itemData[9], 65535) == 0))) then
				--	complete = false;
				--	if (forceCompletion) then
				--		gear[i] = itemLink;
				--	end
				--	gearPartial = not forceCompletion;
				--else
					gear[i] = itemLink;
				--end
--			else
--				gear = nil;
--				complete = false;
--				break; 
			end
		end
		if guid == UnitGUID(unitId) and gear ~= nil then -- Make sure inspectee hasn't changed
			player:SetGear(gear, gearPartial);
		end
	end
end

function CharDB.EventHandlers.RAID_ROSTER_UPDATE()
	CharDB.Debug("Running roster update~");
	local count = GetNumPartyMembers();
	local unitId, guid, player;
	for i=1,count do
		unitId = "party"..i;
		for guid, player in pairs(CharDBData.players) do
			if (not player:ValidateUnitId() and UnitGUID(unitId) == guid) then
				player.unitId = unitId;
			end
		end
		
		CharDB.AddToInspectQueue(unitId);
	end

	count = GetNumRaidMembers();
	for i=1,count do
		unitId = "raid"..i
		for guid, player in pairs(CharDBData.players) do
			if (not player:ValidateUnitId() and UnitGUID(unitId) == guid) then
				player.unitId = unitId;
			end
		end
		CharDB.AddToInspectQueue(unitId);
	end
end

CharDB.EventHandlers.PARTY_MEMBERS_CHANGED = CharDB.EventHandlers.RAID_ROSTER_UPDATE

function CharDB.EventHandlers.OnUpdate(elapsed)
	CharDB.timeSinceLastUpdate = CharDB.timeSinceLastUpdate + elapsed; 	

	if (CharDB.timeSinceLastUpdate > CHARDB_UPDATE_INTERVAL and not CharDB.runningUpdate) then
		CharDB.runningUpdate = true;
		
		CharDB.RunPeriodicTasks(CharDB.timeSinceLastUpdate);
		
		CharDB.timeSinceLastUpdate = 0;
--		CharDB.timeSinceLastUpdate = CharDB.timeSinceLastUpdate - CHARDB_UPDATE_INTERVAL;
		CharDB.runningUpdate = false;
	end
end

function CharDB.GetPlayerByName(name) 
	for _, player in pairs(CharDBData.players) do
		if (player.name == name) then return player end
	end
end

function CharDB.GetPlayerByUnitId(unitId) 
	for _, player in pairs(CharDBData.players) do
		if (player.unitId == unitId) then return player end
	end
end

function CharDB.ParseSimcraftOutput(text) 
	text = gsub(text, '[ 	]+', ' ')
--	text = gsub(text, '[ 	]*\n[ 	]*', '\n')
	local parts = {strsplit(' ', text)}
	local statValues, linesParts, player, stat, value
	
	for i, part in ipairs(parts) do
		local b = (strfind(part, '=') == nil);
		if (not b) then
			if (player) then
				stat, value = strsplit('=',part)
				statValues[stat] = tonumber(value)
			end
		end
		if ((b or i == #parts) and player) then
			player:SetStatValues(statValues);
		end
		if (b) then
			player = CharDB.GetPlayerByName(part) 
			statValues = {}
		end
	end
end

function CharDB.tableequals(table1, table2, path) 
   -- if (path == nil) then path = '' end
	local equal = true
	for tableIndex, table in ipairs({table1, table2}) do
		for name, _ in pairs(table) do
			if (type(table1[name]) ~= type(table2[name])) then
				equal = false;
			   -- Spew(path..'-'..name..'-1', table1[name]);
			   -- Spew(path..'-'..name..'-2', table2[name])
				break;
			elseif (type(table1[name]) == 'table') then
				if (not CharDB.tableequals(table1[name], table2[name], path..'-'..name)) then
					equal = false;
					break;
				end
			elseif (table1[name] ~= table2[name]) then
				equal = false;
			 --   Spew(path..'-'..name..'-1', table1[name]);
			--	Spew(path..'-'..name..'-2', table2[name])
			end 
			--	   if (equal == false) then
			--		   break;
			--	  end
		end
		--	if not equal then break end;
	end
	return equal
end

CharDB.Init();