local function htmlencode(string) 
	return gsub(string, ' ', '+')
end

CHARDB_BLIZZARD2SIMCRAFT_WEAPONSTATS = {
	["DPS_TEMPLATE"]="DPS",
--	["DAMAGE_TEMPLATE_WITH_SCHOOL"]="damage",
--	["DAMAGE_TEMPLATE"]="damage",
	["WEAPON_SPEED"]="Speed",
	["DPS_TEMPLATE"]="Wdps",

}

CHARDB_BLIZZARD2SIMCRAFT_STATS = {
	["ITEM_MOD_AGILITY"]="Agi", 
	["ITEM_MOD_SPIRIT"] = "Spi",
	["ITEM_MOD_STAMINA"] = "Sta", 
	["ITEM_MOD_STRENGTH"] = "Str", 
	["ITEM_MOD_INTELLECT"] = "Int", 

	["ITEM_MOD_SPELL_POWER"] = "SP",
	["ITEM_MOD_MANA_REGENERATION"] = "MP5",
	["ITEM_MOD_SPELL_PENETRATION"] = "SPen",

	["ITEM_MOD_ATTACK_POWER"] = "AP",
	["ITEM_MOD_EXPERTISE_RATING"] = "Exp",
	["ITEM_MOD_ARMOR_PENETRATION_RATING"] = "ArPen",
	
	["ARMOR_TEMPLATE"] = "Armor",
--	["ITEM_MOD_RESILIENCE_RATING"] = "Resil",

--	["ITEM_MOD_PARRY_RATING"] = "Parry",
--	["ITEM_MOD_DODGE_RATING"] = "Dodge",
--	["ITEM_MOD_BLOCK_RATING"] = "Block",
--	["ITEM_MOD_DEFENCE_RATING"] = "Def",

	["ITEM_MOD_CRIT_RATING"] = "Crit",
	["ITEM_MOD_HASTE_RATING"] = "Haste",
	["ITEM_MOD_HIT_RATING"] = "Hit",

}

CHARDB_BLIZZARD2SIMCRAFT_ALLSTATS = {}
for k,v in pairs(CHARDB_BLIZZARD2SIMCRAFT_WEAPONSTATS) do CHARDB_BLIZZARD2SIMCRAFT_ALLSTATS[k]=v end
for k,v in pairs(CHARDB_BLIZZARD2SIMCRAFT_STATS) do CHARDB_BLIZZARD2SIMCRAFT_ALLSTATS[k] = v end

CHARDB_ARMORY_CLASS_IDS = {
	WARRIOR = 1,
	PALADIN = 2,
	HUNTER = 3,
	ROGUE = 4,
	PRIEST = 5,
	DEATHKNIGHT = 6,
	SHAMAN = 7,
	MAGE = 8,
	DRUID = 11
}
CHARDB_SIMCRAFT_SERVERLOCATION = "region=us,server="..htmlencode(GetRealmName()) -- FIXME region

CHARDB_INVENTORY_IDS = {
	[01]='head',
	[02]='neck',
	[03]='shoulder',
	[04]='shirt',
	[05]='chest',
	[06]='waist',
	[07]='legs',
	[08]='feet',
	[09]='wrists',
	[10]='hands',
	[11]='finger1',
	[12]='finger2',
	[13]='trinket1',
	[14]='trinket2',
	[15]='back',
	[16]='main_hand',
	[17]='off_hand',
	[18]='ranged',
	[19]='tabard'
}
CHARDB_ITEMLOC = {
	INVTYPE_2HWEAPON={16,17},
	INVTYPE_BODY=04,
	INVTYPE_CHEST=05,
	INVTYPE_CLOAK=15,
	INVTYPE_FEET=08,
	INVTYPE_FINGER={11,12},
	INVTYPE_HAND=10,
	INVTYPE_HEAD=01,
	INVTYPE_HOLDABLE=17,
	INVTYPE_LEGS=07,
	INVTYPE_NECK=02,
	INVTYPE_RANGED=18,
	INVTYPE_RANGEDRIGHT=18,
	INVTYPE_RELIC=18,
	INVTYPE_ROBE=05,
	INVTYPE_SHIELD=17,
	INVTYPE_SHOULDER=03,
	INVTYPE_TABARD=19,
	INVTYPE_THROWN=18,
	INVTYPE_TRINKET={13,14},
	INVTYPE_WAIST=09,
	INVTYPE_WEAPON={16,17},
	INVTYPE_WEAPONMAINHAND=16,
	INVTYPE_WEAPONOFFHAND=17,
}

local itemData = LibItemDB.GetItemData;

CharDBUnit = {}
CharDBUnit.__index = CharDBUnit;

function CharDBUnit:New(unitId)
	local object = {};
	setmetatable(object, self);
	object.unitId = unitId;
	object.GUID = UnitGUID(unitId);
	object.statValues = {}
	object:RefreshData();
	object:ResetGear();
	object:ResetTalents();
	return object;
end

function CharDBUnit:RefreshData()
	if (self:ValidateUnitId()) then
		self.name = UnitName(self.unitId);
		self.classLocal, self.class = UnitClass(self.unitId);
		self.level = UnitLevel(self.unitId);
	end
end

function CharDBUnit:Complete()
	return self:GearComplete() and self:TalentsComplete();
end

function CharDBUnit:GearComplete()
	return self.gear ~= nil and self.gearPartial ~= true;
end

function CharDBUnit:TalentsComplete()
	return self.talentGroups ~= nil;
end

function CharDBUnit:Outdated(time)
	return self:GearOutdated(time) or self:TalentsOutdated(time);
end

function CharDBUnit:GearOutdated(time)
	return self.gearDate < time;
end

function CharDBUnit:TalentsOutdated(time)
	return self.talentsDate < time;
end

function CharDBUnit:SetGear(gear, gearPartial)
	self:ResetGear();
	self:RefreshData();
	self.gear = gear;
	self.gearPartial = gearPartial;
	self.gearDate = time();
end

function CharDBUnit:ResetGear()
	self.gearPartial = false;
	self.gear = nil;
	self.gearDate = 0;
	self.sets = nil;
	self.metaGem = nil;
end

function CharDBUnit:SetTalents(talentGroups, activeTalentGroup)
	self:ResetTalents();
	self:RefreshData();
	self.activeTalentGroup = activeTalentGroup;
	self.talentGroups = talentGroups;
	self.talentsDate = time();
end

function CharDBUnit:ResetTalents()
	self.talentGroups = nil;
	self.activeTalentGroup = nil;
	self.talentsDate = 0;
end

function CharDBUnit:SetStatValues(statValues, spec) 
	if (spec == nil) then spec = self:GetSpecName() end;
	if (type(spec) == 'number') then spec = self:GetSpecName(spec) end;
	local gemStart, gemEnd = 3446, 3517;
	if (CharDBData.gemType == 'epic') then
		gemStart, gemEnd = 3518, 3590;
	end
	local gemValue = 0;
	local gems = {3446, 3446, 3446, 3446};
	local gemsValue = {-1, -1, -1, -1};
	for gemId=gemStart,gemEnd do
		gemValue = 0;
		gemData = LibItemDB.GetGemData(gemId);
		for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_STATS) do
		    if (statValues[simName] and gemData[blizName]) then 
		        gemValue = gemValue + statValues[simName] * gemData[blizName];
		    end
		end
		for colorId, _ in ipairs(LIBITEMDB_SOCKETCOLORS) do
			if (gemData["GEM_COLORS"][colorId] and gemValue > gemsValue[colorId]) then
				gems[colorId] = gemId;
				gemsValue[colorId] = gemValue;
			end
		end
	end
	for colorId, color in ipairs(LIBITEMDB_SOCKETCOLORS) do
		statValues[color] = gemsValue[colorId];
	end
--	Spew('gems', gems);
	self.statValues[self:GetSpecName()] = statValues;
end

function CharDBUnit:ValidateUnitId() 
	local result = self.unitId ~= nil and UnitGUID(self.unitId) == self.GUID
	if (not result) then self.unitId = nil end
	return result;
end

function CharDBUnit:GetTalentURL() 
	if (not self:TalentsComplete()) then return nil end;
	local url = 'h'..'ttp://www.wowarmory.com/talent-calc.xml?cid='..CHARDB_ARMORY_CLASS_IDS[self.class].."&tal="
	local talents = self:GetTalents();
	for i=1,#talents do
		for j=1, #talents[i] do
			url = url..talents[i][j]
		end
	end
	return url;
end

function CharDBUnit:GetSpec(...) 
	if (not self:TalentsComplete()) then return nil end;
	local talentsGroup = self.activeTalentGroup
	if (select("#", ...) > 0) then
		talentsGroup = select(1, ...)
	end	local spec = {}
	local talents = self:GetTalents(talentsGroup)
	for i=1,3 do
		spec[i] = 0
		for j=1,#talents[i] do
			spec[i] = spec[i] + talents[i][j]
		end
	end
	return spec;
end

function CharDBUnit:GetMetaGem()
	if (not self:GearComplete()) then return nil end;
	if (self.metaGem == nil) then
		local helm = itemData(self.gear[1]);
		--assert(#helm.GEMS == 4)
		--for i=1,4 do
		for _, gemId in pairs(helm.GEMS) do
--			local gemId = helm.gemIds[i]
			for metaGemId, metaGemName in pairs(CHARDB_METAGEMS) do
				if (gemId == metaGemId) then
					self.metaGem = metaGemName;
				end
			end
		end
	end
	return self.metaGem
end

function CharDBUnit:GetSets()
	if (not self:GearComplete()) then return nil end;
	if (self.setBonuses == nil) then
		local setData = CHARDB_SETS[self.class]
		local sets = {}
		for setName, set in pairs(setData) do
			sets[setName] = 0;
			for _, item in pairs(self.gear) do
				for _, itemID in pairs(set) do
					if (itemID == itemData(item).itemId) then
						sets[setName] = sets[setName] + 1;
					end
				end
			end
		end
		self.sets = sets;
	end
	return self.sets
end

function CharDBUnit:GetTalents(...)
	if (not self:TalentsComplete()) then return nil end;
	local talentsGroup = self.activeTalentGroup
	if (select("#", ...) > 0) then
		talentsGroup = select(1, ...)
	end
	return self.talentGroups[talentsGroup];
end

function CharDBUnit:GetSpecName(...) 
	if (not self:TalentsComplete()) then return nil end;
	local talentsGroup = self.activeTalentGroup
	if (select("#", ...) > 0) then
		talentsGroup = select(1, ...)
	end
	local spec = {}	
	local max = 0;
	local maxIndex = 0;
	local spec = self:GetSpec(talentsGroup)
	if (spec == nil) then return nil end;
	for tabIndex, count in pairs(spec) do
		if (count > max) then
			maxIndex = tabIndex
			max = count
		end
	end
	if (max == 0) then
		return ''
	else
		return CHARDB_TALENT_TAB_NAMES[self.class][maxIndex];
	end
end

function CharDBUnit:GetSimcraftParams()

	-- Do basic data -- this is a armory player option!
	local text = "player="..self.name..","..CHARDB_SIMCRAFT_SERVERLOCATION..',talents='..self:GetSpecName()..'\n'
	text = text.."level="..self.level..'\n';
	text = text.."talents="..self:GetTalentURL() ..'\n';
	
	-- Do gear
	for i, link in pairs(self.gear) do
		text=text..CHARDB_INVENTORY_IDS[i]..'='..self.GetItemString(link)..'\n';
	end
	
	-- Do set bonuses
	local setBonuses = self:GetSets();
	for setName, count in pairs(setBonuses) do
		if (count >= 2) then
			text = text..setName.."_2pc=1"..'\n'
			if (count >= 4) then
				text = text..setName.."_4pc=1"..'\n'
			else
				text = text..setName.."_4pc=0"..'\n'
			end
		else
			text = text..setName.."_2pc=0"..'\n'
			text = text..setName.."_4pc=0"..'\n'
		end
	end
	
	-- Do meta gem
	local metaGem = self:GetMetaGem()
	if (metaGem ~= nil) then
		text = text.."meta_gem="..metaGem..'\n'
	end
	
	-- Do "unique" item (items with proc or on use effects)
	local uniqueItems = {}
	for _, itemLink in pairs(self.gear) do
		local name = SimcraftizeName(itemData(itemLink).name)
		for _, uniqueItemName in pairs(CHARDB_UNIQUEITEMS) do
			if (name == CHARDB_UNIQUEITEMS) then
				tinsert(uniqueItems, name)
			end
		end
	end
	if (#uniqueItems > 0) then
		local uniqueItemsTest = strjoin('/',  unpack(uniqueItems))
		text = text..'items='..uniqueItemsTest..'\n'
	end

	return text;
end

function CharDBUnit:CompareItem(link, spec) 
	local compareItems = self:ItemInSlot(link, spec)
	local linkValue = self:RateItem(link, spec);
	if (linkValue == 0) then return {} end;
	local result = {}
	for _, item in ipairs(compareItems) do
		result[item] = self:RateItem(item, spec)-linkValue;
	end
	return result;
end

function CharDBUnit:ItemInSlot(link, spec)
	if (not self:GearComplete()) then return nil end;
	local loc = CHARDB_ITEMLOC[select(9, GetItemInfo(link))]
	if (type(loc) == 'table') then
		assert(#loc == 2)
		return {self.gear[loc[1]], self.gear[loc[2]]}
	else
		return {self.gear[loc]}
	end
end

function CharDBUnit:RateItem(link, spec) 
	if (link == nil) then return 0 end;
	if (spec == nil) then spec = self:GetSpecName() end;
	if (type(spec) == 'number') then spec = self:GetSpecName(spec) end;
	if (self.statValues[spec] == nil) then return nil end;
	local cleanItemString = LibItemDB.RemoveEnchantsFromItemString(LibItemDB.LinkToItemString(link));
	local statValues = self.statValues[spec];
	local data = itemData(cleanItemString);
	local result = 0;

	for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_ALLSTATS) do
	    if (statValues[simName] and data[blizName]) then 
--			CharDB.Debug(format('%s: %.1f * %.1f = %.1f', simName, statValues[simName], data[blizName], statValues[simName]*data[blizName]));
	        result = result + statValues[simName] * data[blizName];
	    end
	end
	
--	CharDB.Debug(format("Stats: %.1f", result));
	
	if (data['EMPTY_SOCKET']) then
	
		local correctSocketingValue = 0;
		assert(data["ITEM_SOCKET_BONUS_GREY"] ~= nil)
	    for stat, value in pairs(data["ITEM_SOCKET_BONUS_GREY"]) do
	        for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_ALLSTATS) do
	            if (stat == blizName and statValues[simName]) then 
	                correctSocketingValue = correctSocketingValue + statValues[simName] * value;
	            end
	        end
	    end
		
		local incorrectSocketingValue = 0
		local bestGemValue = 0;
		for colorId, colorName in ipairs(LIBITEMDB_SOCKETCOLORS) do
			if (statValues[colorName] and bestGemValue < statValues[colorName]) then
				bestGemValue = statValues[colorName];
			end
		end

		local sockets = data['EMPTY_SOCKET']
	    if (not sockets.array) then
	    	sockets = {sockets}
	    end
        for _, color in ipairs(sockets) do
            if (statValues[color]) then
                correctSocketingValue = correctSocketingValue + statValues[color];
	            if (color == 'META') then
		            incorrectSocketingValue = incorrectSocketingValue + statValues[color];
		        else
		            incorrectSocketingValue = incorrectSocketingValue + bestGemValue;
		        end
            end
        end
	    
--		CharDB.Debug(format("correctSocketingValue: %d", correctSocketingValue));
--	    CharDB.Debug(format("incorrectSocketingValue: %d", incorrectSocketingValue));

	    result = result + (incorrectSocketingValue > correctSocketingValue and incorrectSocketingValue or correctSocketingValue);
	end
--	CharDB.Debug(format("%s: %.1f", link, result));
	return result;
end

function CharDBUnit.GetItemString(link) 
	local item = itemData(link);
	local string = SimcraftizeName(item['INFO'][1]);
	--if (#item.stats > 0) then
	local statString = "";
	for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_STATS) do
		if (item[blizName]) then 
			statString = statString..item[blizName]..simName..'_';
		end
	end
	if (strlen(statString) ~= 0) then
		statString = strsub(statString, 0, strlen(statString)-1);
		string = string..",stats="..statString;
	end
	
	if(item["INVTYPE_RANGED"] or item["INVTYPE_2HWEAPON"] or item["INVTYPE_WEAPON"] or item["INVTYPE_WEAPONMAINHAND"] or item["INVTYPE_WEAPONOFFHAND"]) then
		--assert(#item.weaponStats == 2)
		local subtype = LIBITEMDB_ITEMTYPES[item['INFO'][7]];
		if (subtype ~= 'Wand' and subtype ~= nil) then
			subtype = SimcraftizeName(subtype);
			subtype = gsub(subtype, 'twohanded_', '');
			subtype = gsub(subtype, 'onehanded_', '');
			string = string..",weapon="..subtype..'_';
		
			for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_WEAPONSTATS) do
				if (item[blizName]) then 
					string = string..item[blizName]..simName..'_';
				end
			end
			string = strsub(string, 0, strlen(string)-1);
		end
	end
	
	if (item["ENCHANT"]) then
		local enchantData = LibItemDB.GetEnchantData(item["ENCHANT"]);
		local enchantString = "";
		for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_STATS) do
			if (enchantData[blizName]) then 
				enchantString = enchantString..enchantData[blizName]..simName..'_';
			end
		end
		if (strlen(enchantString) ~= 0) then
			string = string..",enchant="..strsub(enchantString, 0, strlen(enchantString)-1);
		end
	end

	
	if (item["GEMS"] and #item["GEMS"] ~= 0) then
		
		local gemsString = ""
		local gemsData = {};
		for i,gem in pairs(item["GEMS"]) do
			if (CHARDB_METAGEMS[gem]) then
				gemsString = gemsString..CHARDB_METAGEMS[gem]..'_'
			end
			local gemData = LibItemDB.GetGemData(gem);
			for name, value in pairs(gemData) do
				if (type(value) == 'number') then
					if (not gemsData[name]) then gemsData[name] = 0 end
					gemsData[name] = gemsData[name] + value;
				end
			end
		end
		
		if (item["ITEM_SOCKET_BONUS"]) then
			for name, value in pairs(item["ITEM_SOCKET_BONUS"]) do
				if (type(value) == 'number') then
					if (not gemsData[name]) then gemsData[name] = 0 end
					gemsData[name] = gemsData[name] + value;
				end
			end
		end
		for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_STATS) do
			if (gemsData[blizName]) then 
				gemsString = gemsString..gemsData[blizName]..simName..'_';
			end
		end
		if (strlen(gemsString) ~= 0) then
			string = string..",gems="..strsub(gemsString, 0, strlen(gemsString)-1);
		end
	end

	
	return string
end

function SimcraftizeName(name) 
	if (name == nil) then return nil end
	return gsub(gsub(string.lower(name), "[^a-z ]", ""), " +", "_")
end